from unittest import TestCase

import codecs
import os
import genderator

path = os.path.dirname(__file__)

DELTA = 0.01
MAX_PERCENTAGE_ERROR = 0.05
TEST_FILE_LINES = 100000


class TestParser(TestCase):
    __data = {}
    __parser = genderator.Parser()

    def remove_file_comments(self, relative_path):
        """
        Generator to remove comments from a file.

        Params:
            file: File to be processed.
        """
        with codecs.open(os.path.join(path, 'data', relative_path), 'r', 'UTF-8') as file:
            for line in file:
                line = line.strip()
                if not line.startswith('#'):
                    yield line

    def test_gender_guessing(self):
        """
        Check if mistake percentage is lower than MAX_ERROR_PERCENTAGE while guessing gender.
        """
        mistakes = 0
        for line in self.remove_file_comments('valid_names.tsv'):
            (name, first_surname, second_surname, male_probability) = line.split('\t')
            fullname = ' '.join([name, first_surname, second_surname])
            answer = self.__parser.guess_gender(fullname)
            confidence = float(answer['confidence'])
            calculated_male_probability = confidence if answer['gender'] == 'Male' else 1 - confidence

            if abs(calculated_male_probability - float(male_probability)) > DELTA:
                mistakes += 1
            self.assertLess(mistakes / TEST_FILE_LINES * 100, MAX_PERCENTAGE_ERROR,
                            'Mistakes percentage greater than ' + str(MAX_PERCENTAGE_ERROR))

    def test_match(self):
        """
        Check if there is a match for every full name in the test file.
        """
        for line in self.remove_file_comments('valid_names.tsv'):
            (name, first_surname, second_surname, male_probability) = line.split('\t')
            fullname = ' '.join([name, first_surname, second_surname])
            answer = self.__parser.guess_gender(fullname)
            self.assertIsNotNone(answer, 'Fullname doesn\'t match: ' + fullname)

    def test_name_guessing(self):
        """
        Check if mistake percentage is lower than MAX_ERROR_PERCENTAGE while guessing name.
        """
        mistakes = 0
        for line in self.remove_file_comments('valid_names.tsv'):
            (name, first_surname, second_surname, male_probability) = line.split('\t')
            fullname = ' '.join([name, first_surname, second_surname])
            answer = self.__parser.guess_gender(fullname)
            try:
                if answer['real_name'] != name:
                    mistakes += 1
                self.assertLess(mistakes / TEST_FILE_LINES * 100, MAX_PERCENTAGE_ERROR,
                                'Mistakes percentage greater than ' + str(MAX_PERCENTAGE_ERROR))
            except TypeError:
                print(fullname)


import unicodedata

good_accents = {
    u'\N{LATIN CAPITAL LETTER N WITH TILDE}',
    u'\N{LATIN SMALL LETTER N WITH TILDE}',
    u'\N{LATIN CAPITAL LETTER C WITH CEDILLA}',
    u'\N{LATIN SMALL LETTER C WITH CEDILLA}'
}


class Normalizer:

    def normalize(text, options={}):
        """
        Normalize a given text applying all normalizations.

        Params:
            text: The text to be processed.

        Returns:
            The text normalized.
        """
        if options.get('whitespaces', True):
            text = Normalizer.remove_extra_whitespaces(text)

        if options.get('hyphens', True):
            text = Normalizer.replace_hyphens(text)

        accents = options.get('hyphens', True)
        symbols = options.get('symbols', True)

        if accents and symbols:
            text = Normalizer.normalize_unicode(text)
        elif accents:
            text = Normalizer.remove_accent_marks(text)
        else:
            text = Normalizer.remove_symbols(text)

        return text.lower()

    @staticmethod
    def normalize_unicode(text):
        """
        Remove accent marks and symbols from input text.

        This function has the same effect that remove_accent_marks and remove_symbols but use a single loop.

        Params:
            text: The text to be processed.

        Returns:
            The text without accent marks and symbols.
        """
        categories = ['Mn', 'Sc', 'Sk', 'Sm', 'So']

        return ''.join(c for c in unicodedata.normalize('NFKC', text)
                       if unicodedata.category(c) not in categories or c in good_accents)

    @staticmethod
    def remove_accent_marks(text):
        """
        Remove accent marks from input text.

        Params:
            text: The text to be processed.

        Returns:
            The text without accent marks.
        """
        return ''.join(c for c in unicodedata.normalize('NFKC', text)
                       if unicodedata.category(c) != 'Mn' or c in good_accents)

    @staticmethod
    def remove_extra_whitespaces(text):
        """
        Remove extra whitespaces from input text.

        This function removes whitespaces from the beginning and the end of
        the string, but also duplicated whitespaces between words.

        Params:
            text: The text to be processed.

        Returns:
            The text without extra whitespaces.
        """
        return ' '.join(text.strip().split())

    @staticmethod
    def replace_hyphens(text):
        """
        Remove hyphens from input text.

        Params:
            text: The text to be processed.

        Returns:
            The text without hyphens.
        """
        return text.replace('-', ' ')

    @staticmethod
    def remove_symbols(text):
        """
        Remove symbols from input text.

        Params:
            text: The text to be processed.

        Returns:
            The text without symbols.
        """
        return ''.join(c for c in unicodedata.normalize('NFKC', text)
                       if unicodedata.category(c) != 'Mn' or c in good_accents)



#!/user/bin/env python

import re

from setuptools import setup, find_packages


version = re.search(
    '^__version__\s*=\s*\'(.*)\'',
    open('genderator/__init__.py').read(),
    re.M).group(1)

setup(name='genderator',
      version=version,
      description='Python library to guess gender given a spanish full name',
      author='David Moreno-Garcia',
      author_email='david.mogar@gmail.com',
      license='MIT',
      url='https://github.com/davidmogar/genderator',
      download_url='https://github.com/davidmogar/genderator/tarball/' + version,
      keywords=['gender', 'guess', 'spanish', 'name'],
      packages=find_packages(exclude=['tests']),
      include_package_data=True,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Topic :: Software Development :: Libraries',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
      ]
      )



import codecs
import os

from collections import OrderedDict
from .utils import Normalizer

path = os.path.dirname(__file__)


class Parser:
    """
    This class offers a simple way to parse Spanish name (from Spain), classifying
    names and surnames and detecting the given name gender.

    Attributes:
        force_combinations (boolean): Force combinations during classification.
        force_split (boolean): Force name split if no surnames detected.
        normalize (boolean): Enable or disable normalization.
        normalizer_options (dict): Normalizer options to be applied.
    """
    __names, __ratios = {}, {}
    __surnames = []

    def __init__(self, force_combinations=True, force_split=True, normalize=True, normalizer_options={}):
        self.__force_combinations = force_combinations
        self.__force_split = force_split
        self.__normalize = normalize
        self.__normalizer_options = normalizer_options

        self._load_data()

    def _load_data(self):
        """
        Load all data files into memory.
        """
        self._load_names()
        self._load_name_surname_ratios()
        self._load_surnames()

    def _load_names(self):
        """
        Load names data file.

        This file contains a list of spanish given names with the probability for
        each one to be a male or female name.
        """
        for line in self.remove_file_comments('names_ine.tsv'):
                (name, frequency, prob_male) = line.split('\t')
                self.__names[name] = float(prob_male)

    def _load_name_surname_ratios(self):
        """
        Load name/surnames ratios data file.

        The file contains a list of names and surnames with the probability for each
        one to be a name (lower values) or a surname (higher values).
        """
        for line in self.remove_file_comments('name_surname_ratio.tsv'):
            (key, val) = line.split('\t')
            self.__ratios[key] = float(val)

    def _load_surnames(self):
        """
        Load names data file.

        This file contains a list of spanish surnames.
        """
        for line in self.remove_file_comments('surnames_ine.tsv'):
            self.__surnames.append(line.split('\t')[0])

    def remove_file_comments(self, relative_path):
        """
        Generator to remove comments from a file.

        Params:
            file: File to be processed.
        """
        with codecs.open(os.path.join(path, 'data', relative_path), 'r', 'UTF-8') as file:
            for line in file:
                line = line.strip()
                if not line.startswith('#'):
                    yield line

    def guess_gender(self, fullname):
        """
        Guess the gender of the given full name.

        Params:
            fullname: Full name from where we want to guess the gender.

        Returns:
            A JSON string with all the computed information.
        """
        if isinstance(fullname, str):
            if self.__normalize:
                fullname = Normalizer.normalize(fullname, self.__normalizer_options)

            names, surnames = self._classify(fullname)

            if names and (surnames or (self.__force_split and self._is_splittable(names))):
                real_name, ratio = self._get_gender_ratio(list(names))
                return self._create_answer(real_name, ratio, names, surnames)

    def _is_splittable(self, names):
        """
        Check if a list of names can be splitted in names and surnames.

        Params:
            names: List of name to be checked.

        Returns:
            True if can be splitted or false otherwise.
        """
        names_iterator = iter(names)
        next(names_iterator)
        for name in names_iterator:
            if name in self.__ratios and self.__ratios[name] < 1: return True
        return False

    def _classify(self, fullname):
        """
        Split fullname into tokens and classify them into names and surnames based on datasets.

        Params:
            fullname: Full name to be classified.

        Returns:
            Two lists, one with names and the other with surnames.
        """
        names, surnames = [], []
        unclassified = []
        processed = []

        for word in fullname.split():
            combination_result = None

            if self.__force_combinations:
                combination_result = self._combine_words(processed, word, names)

            if combination_result is None:
                keep_going = True
                if unclassified:
                    if self._classify_word(unclassified[-1] + ' ' + word, names, surnames, unclassified):
                        keep_going = False
                if keep_going:
                    if unclassified:
                        self._classify_word(word, names, surnames)
                    else:
                        self._classify_word(word, names, surnames, unclassified)

                processed.append(word)

        return names, surnames

    def _combine_words(self, processed, word, names):
        """
        Try to combine last processed word with the word received as parameter.

        If the combination of both words is a name, this is added to the list, replacing
        the name added previously.

        Params:
            processed: List of words already processed.
            word: Current word.
            names: List of classified names.

        Returns:
            A valid combination if found or None otherwise.
        """
        if processed and names:
            last_word = processed[-1]
            if last_word == names[-1]:
                combination = last_word + ' ' + word
                if combination in self.__names:
                    names.pop(names.index(last_word))
                    names.append(combination)
                    processed[-1] = combination
                    return combination
        return None

    def _classify_word(self, word, names, surnames, unclassified=None):
        """
        Try to classify a word in name or surname based on datasets.

        Params:
            word: Word to be classified.
            names: List of classified names.
            surnames: List of classified surnames.
            unclassified: List of words without match.

        Returns:
            True if the word was classified. False otherwise.
        """
        classified = True

        if word in self.__ratios:
            if (not names or self.__ratios[word] > 0.5) and not surnames:
                names.append(word)
            else:
                surnames.append(word)
        else:
            if word in self.__surnames and names:
                surnames.append(word)
            elif word in self.__names and not surnames:
                names.append(word)
            else:
                if unclassified is not None:
                    unclassified.append(word)
                classified = False

        if classified and unclassified is not None: unclassified.clear()

        return classified

    def _get_gender_ratio(self, names):
        """
        Returns the male/female ratio for the given names.

        To do this, the function compute possible names combining items on the list
        and try to form the longest name possible.

        The value returned go from 0 to 1. Values near to 1 represent a higher possibility
        of the evaluated name to be a male name.

        Params:
            names: List of names.

        Returns:
            The longest name computed by combining items in the list,
            and the male/female ratio.
        """
        for i in range(len(names), 0, -1):
            real_name = ' '.join(names[:i])
            if real_name in self.__names:
                return real_name, self.__names[real_name]

    def _create_answer(self, real_name, ratio, names, surnames):
        """
        Process computed data and generated a JSON answer.

        Params:
            real_name: Real name (computed name) extracted from the original text.
            ratio: Male/female ratio.
            names: Names identified on the original text.
            surnames: Surnames identified on the original text.

        Returns:
            A JSON string with all the computed information.
        """
        answer = OrderedDict()
        answer['names'] = names
        answer['surnames'] = surnames
        answer['real_name'] = real_name
        male = ratio > 0.5
        answer['gender'] = 'Male' if male else 'Female'
        answer['confidence'] = ratio if male else 1 - ratio
        return answer